import de.janicke.hj.tools.io.FileCopy;
import de.janicke.hj.tools.io.MyTokenizer;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.*;

/** Backup is a small class that looks for updated files in a directory.
  * To check if a File is updated the last modification date is checked against*
  * the lastcheck date.
  * All settings are read from a property file.
  * You may specify endings for files like "suffix = java" to specify that java
  * source files should be checked.
  */
public class Backup implements ActionListener {

	/** HashMap containing the observed File as key and the autocreated
	  * backup directory in form %out%/%observe%/date as object.
	  */
	private HashMap observed = new HashMap();

	/** HashMap containing the suffixes to backup if newer than the time of
	  * the last check.
	  */
	private HashMap suffix = new HashMap();

	/** The output folder for backup. This folder is created at startup. */
	private File outputFolder = null;

	/** Stores the time of the last backup. */
	private long lastBackup = 0;

	/** The default backup intervall. */
	private int backupInterval = 0;

	/** The Timer used to backup in intervalls. */
	private javax.swing.Timer timer;

	/** Constructs a new Backup class. */
	public Backup(String fname) {
		init(fname);
	}

	private PrintStream out = System.out;
	private PrintStream err = System.err;

	/** Starts the autobackup. A check is made once every intervall
	  * milliseconds.
	  */
	public void start() {
		timer.start();
		out.println("autobackup started!");
	}

	/** Stops the autobackup. */
	public void stop() {
		timer.stop();
		out.println("autobackup stopped!");
	}

	/** Initializes this Backup.
	  * The properties are read from the property file given at startup.
	  * If out is not specified, a folder "default_backup" in the current
	  * directory is made. The program runs if no observed Folder or no
	  * suffixes are set.
	  * @param fname The path of the property file.
	  */
	private void init(String fname) {
		StringBuffer sets = new StringBuffer();
		String line;

		try {
			BufferedReader fr = new BufferedReader(new FileReader(fname));
			while((line = fr.readLine()) != null) {
				sets.append(line);
				sets.append("\n");
			}
		} catch (Exception exc) {
			exc.printStackTrace();
			System.exit(1);
		}

		MyTokenizer mt = new MyTokenizer(sets.toString());
		mt.setSkipComments(true);

		String lastWord = "";
		while (mt.next() != MyTokenizer.EOF) {
			switch (mt.type) {
				case MyTokenizer.WORD:
					lastWord = (String) mt.value;
					break;
				case MyTokenizer.NUMBER:
					if (lastWord.equalsIgnoreCase("time")) {
						backupInterval = ((Number) mt.value).intValue();
					}
					lastWord = "";
					break;
				case '\"':
					if (lastWord.equalsIgnoreCase("out")) {
						outputFolder = new File(mt.value.toString());
					}
					if (lastWord.equalsIgnoreCase("observe")) {
						File observedFile = new File(mt.value.toString());
						observed.put(observedFile, new File(outputFolder, observedFile.getName()));
					}
					if (lastWord.equalsIgnoreCase("suffix")) {
						suffix.put(mt.value.toString(), null);
					}
					lastWord = "";
					break;
				default:
			}
		}

		if (outputFolder == null) {
			outputFolder = new File("default_backup");
			out.println("no outputfile specified, using: " + outputFolder);
		}
		outputFolder.mkdirs();

		if (backupInterval == 0) {
			out.println("No time specified using 1800000 (30 min) as default");
			backupInterval = 1800000;
		}
		lastBackup = System.currentTimeMillis();
		timer = new javax.swing.Timer(backupInterval, this);
		Iterator i = observed.values().iterator();

		// creating a output dir for each observed folder.
		out.println("Creating directories for each observed directory...");
		while (i.hasNext()) {
			File f = (File) i.next();
			out.println("creating output directory: " +  f.getAbsolutePath());
			f.mkdir();
		}
	}

	public void actionPerformed(ActionEvent e) {
		checkNow();
	}

	public void setDelay(int delay) {
		timer.setDelay(delay);
	}

	public String makeFName(long time) {
		Date d = new Date(time);
		StringBuffer s = new StringBuffer(d.toString());
		for (int i=0; i<s.length(); i++) {
			char c = s.charAt(i);
			switch(c){
				case ':' :
				case ';' :
				case ' ' :
					s.setCharAt(i,'_');
					break;
				case '/' :
				case '\\':
					s.setCharAt(i,'-');
					break;
				default:
			}
		}
		return s.toString();
	}

	public synchronized void checkNow() {
		// check
		long now = System.currentTimeMillis();
		Iterator i = observed.keySet().iterator();
		File folder;
		File output;
		String subfoldername = makeFName(now);

		int updated = 0;
		int updatetotal = 0;
		// checking each of observed folders.
		while(i.hasNext()) {
			folder = (File) i.next();
			if (folder.exists()){
				output = (File) observed.get(folder);
				output = new File(output, subfoldername);
				output.mkdir();
				updated = recCopy(lastBackup, folder, output);
				if (updated == 0) {
					output.delete();
				} else {
					out.println("\nbackuped " + updated + " files from " + folder);
				}
				updatetotal += updated;
			} else {
				out.println("Observed folder " + folder.getAbsolutePath() + " has been deleted, moved or renamed!!!\n Removing Folder from observed Folders list.");
				observed.remove(folder);
			}
		}

		out.println("\nlast backup was at " + new Date(now) + ((updatetotal > 0) ? ("\nbacked up " + updatetotal + " files to subfolder: " + subfoldername) : " not neccassary."));
		System.out.print("Enter command [\"help\" for help]: ");

		lastBackup = now;
	}

	public int recCopy(long t, File folder, File output) {
		File[] files = folder.listFiles();

		int cnt = 0;
		for (int i=0; i<files.length; i++) {
			if (files[i].isDirectory()) {
				File subFolder = new File(output, files[i].getName());
				subFolder.mkdir();
				int updated = recCopy(t, files[i], subFolder);
				if (updated == 0) subFolder.delete();
				else cnt += updated;
			}
			else if ((files[i].lastModified() > t || complete)) {
				int sufpos = files[i].getName().lastIndexOf('.');
				String suff = "";
				if (sufpos < 0) {
				} else {
					suff = files[i].getName().substring(files[i].getName().lastIndexOf('.'));
				}

				if (suffix.containsKey(suff) || complete) {
					// Is in suffixlist or complete backup.
					out.print("\nfile " + files[i] + "\n  needs backup...");
					// copy file.
					try {
						FileCopy.copy(files[i], new File(output, files[i].getName()), false);
						cnt++;
						out.print(" done!");
					} catch (IOException exc) {
						out.print(" FAILED!");
						out.println(				"\n******************************************************************* reason is");
						exc.printStackTrace();
						out.println(				"************************************************************************* end");
					}
				}
			} else {
			}
		}
		return cnt;
	}

	public void createZip() {
	}

	boolean complete = false;
	public synchronized void complete() {
		stop();
		complete = true;
		Iterator i = observed.keySet().iterator();
		File folder;
		File output;
		long now = System.currentTimeMillis();

		while(i.hasNext()) {
			folder = (File) i.next();
			out.println("  complete backup of: " + folder);
			output = (File) observed.get(folder);
			output = new File(output, makeFName(now) + "_complete");
			output.mkdir();
			out.println("    ...storing in: " + output.getAbsolutePath());
			recCopy(lastBackup, folder, output);
			out.println("\n    ...completed!");
		}
		complete = false;
		start();
	}

	public static void main(String args[]) {
		BufferedReader in = new BufferedReader(new InputStreamReader( System.in));

		if (args.length < 1) {
			System.out.println("USAGE: Backup PropertyFile");
			System.exit(1);
		}

		Backup b = new Backup(args[0]);
		b.checkNow();
		b.start();

		String com = "";
		while(true) {
			try {
				System.out.print("Enter command [\"help\" for help]: ");
				com = in.readLine();
				if (com.equalsIgnoreCase("quit")) {
					System.out.println("bye");
					break;
				} else
				if (com.equalsIgnoreCase("start")) {
					b.start();
				} else
				if (com.equalsIgnoreCase("stop")) {
					b.stop();
				} else
				if (com.equalsIgnoreCase("now")) {
					b.checkNow();
				} else
				if (com.equals("help")) {
					System.out.println("\n___HELP___\nquit\t\tquits this program\nstart\t\tstarts autobackup\nstop\t\tstops autobackup\ntime (millis)\tsets autobackup interval\ncomplete\t\tfull backup\nnow\t\tbackup now!\nhelp\t\tthis message\n");
				} else
				if (com.startsWith("time")) {
					com = com.substring(4).trim();
					int inter = Integer.parseInt(com);
					b.setDelay(inter);
				} else
				if (com.equalsIgnoreCase("complete")) {
					System.out.println("complete backup running... This may take a while");
					b.complete();
					System.out.println("complete backup finished for each observed folder.");
				} else {
					System.out.println("unknown command");
				}
			} catch(Exception exc) {
				exc.printStackTrace();
			}
		}

	}
}